@@ -7,12 +7,10 @@ After installation, you can enable it by adding::
 
 in your `jupyterhub_config.py` file.
 """
-
 # We export KubeSpawner specifically here. This simplifies import for users.
 # Users can simply import kubespawner.KubeSpawner in their applications
 # instead of the more verbose import kubespawner.spawner.KubeSpawner.
-
 from kubespawner.spawner import KubeSpawner
 
-__version__ = '0.15.0'
-__all__ = [KubeSpawner]
+__version__ = '0.16.0'
+__all__ = ["KubeSpawner"]
